#include ".\physics.h"
#include ".\ode\ode.h"
#include ".\ode\dCylinder.cpp" 
#include <iostream>
using namespace std;

Physics::Physics(void)
{
	if(!(init()))
		cerr << "Physics engine initialization failed" << endl;
}

Physics::~Physics(void)
{
  dSpaceDestroy (m_space);
  dWorldDestroy (m_world);
}


bool Physics::init()
{
	// create world	
	m_world = dWorldCreate();
	m_space = dHashSpaceCreate(0);
   	setGravity(0.0f,-8.0f,-6.0f);
//	dWorldSetContactMaxCorrectingVel (m_world,0.1);
//	dWorldSetContactSurfaceLayer (m_world,0.001);


	m_contactgroup = dJointGroupCreate(0);

	dWorldSetCFM(m_world,1e-5);
	
	
	// bounding box for the flipper
	//dCreateBox (m_space, 2, 2, 2);

	// board dimensions
	float m_length=0.2f;
	float m_width=0.2f;
	float m_height=0.2f;

	// floor
	dCreatePlane(m_space,0,0,1,0);		// stefan: we can change the angle of the plane by modifiying 
										// the parameters of the plane equation: a*x+b*y+c*z = d
	//ceiling
	dCreatePlane(m_space,0,0,-1,-m_height);

	// top
	dCreatePlane(m_space,0,-1,0,-m_length/2.0);

	// right
	dCreatePlane(m_space,-1,0,0,-m_width/2.0);

	// left
	dCreatePlane(m_space,1,0,0,-m_width/2.0);

	// bottom
	dCreatePlane(m_space,0,1,0,-m_length/2.0);


	// box-left
	//dCreatePlane(m_space,0.644217687,0.764842187,0,-0.076321);
	dCreatePlane(m_space,0.764842187,0.644217687,0,-0.076321);



	// box-right
//	dCreatePlane(m_space,-0.644217687,0.764842187,0,-0.076321);
	dCreatePlane(m_space,-0.764842187,0.644217687,0,-0.076321);

	resetBumperHits();

	return true;
}


dJointID Physics::createFixedJoin(void){
	return dJointCreateFixed(m_world, 0);
}

dJointID Physics::createHingeJoin(void){
	return dJointCreateHinge(m_world, 0);
}

void Physics::resetBumperHits(){
	bumperHits = 0;
}


void Physics::increaseBumperHits(){
	bumperHits++;
}

int Physics::getBumperHits(){
	return bumperHits;
}

// simulation loop
void Physics::simLoop (int run)
{
	// check for collisions, create contact joints
	//dSpaceCollide (m_space,0,&Physics::staticNearCallBack);
  	dSpaceCollide (m_space,this, collisionCallback);

	if (run)
		dWorldQuickStep (m_world,0.01);
	

  // remove all contact joints
  dJointGroupEmpty (m_contactgroup);

}


//-----------------------------------------------------------------
//	Gravity
//-----------------------------------------------------------------

void Physics::setGravity(double x, double y, double z)
{
	m_gravity[0] = x;
	m_gravity[1] = y;
	m_gravity[2] = z;

	dWorldSetGravity(m_world,m_gravity[0],m_gravity[1],m_gravity[2]);
}

double* Physics::getGravity()
{
	return m_gravity;
}



//-------------------------------------------------------------------
//	Creates a new body inside our world, returns body ID
//-------------------------------------------------------------------

dBodyID Physics::createBody()
{
	return dBodyCreate(m_world);
}

//-------------------------------------------------------------------
//	Create a sphere, return the geomID
//-------------------------------------------------------------------

dGeomID Physics::createSphere(float radius)
{	
	return dCreateSphere(m_space,radius);    
}

//-------------------------------------------------------------------
//	Create a box, return the geomID
//-------------------------------------------------------------------

dGeomID	Physics::createBox(float length, float width, float height)
{
	return dCreateBox(m_space, length, width, height);
}


//-------------------------------------------------------------------
//	Create a capped cylinder, return the geomID
//-------------------------------------------------------------------

dGeomID	Physics::createCCyl(float radius, float length)
{
	return dCreateCCylinder (m_space, radius, length);
}


dGeomID Physics::createCyl(float radius, float length){
	return dCreateCylinder (m_space, radius, length);
}


void Physics::setSound(Sound *sound){
	m_sound = sound;
}

//-----------------------------------------------------
//	Create plane, return geomID
//-----------------------------------------------------

dGeomID	Physics::createPlane(dReal a, dReal b, dReal c, dReal d)
{
	return dCreatePlane(m_space,a,b,c,d);
}



void Physics::collisionCallback (void *pData, dGeomID o1, dGeomID o2)
{
	// pData is the "this" that was passed into dSpaceCollide
	Physics* pThis = (Physics*) pData;
	if (pThis)
		pThis->collisionCallback (o1, o2);
}

void Physics::setBallGID(dGeomID ball){
	m_ball = ball;
}


//-------------------------------------------------------------------
//	Call back function called when two bodies are near collision
//-------------------------------------------------------------------

void Physics::collisionCallback (dGeomID o1, dGeomID o2)
{
  int i;


  // only look at collisions with the ball
  if(((m_ball==o1)||(m_ball==o2))||((dGeomGetClass(o1)==dCCylinderClass)||(dGeomGetClass(o2)==dCCylinderClass))){
//	if(!((dGeomGetClass(o1)==dSphereClass)||(dGeomGetClass(o2)==dSphereClass))) return;

  // if (o1->body && o2->body) return;

  // exit without doing anything if the two bodies are connected by a joint
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);


  if (b1 && b2 && dAreConnectedExcluding (b1,b2,dJointTypeContact)) return;

  dGeomID object=0;
  bool ballIsInvolved=false;
	// play colliding sound
  if(dGeomGetClass(o1)==dSphereClass){
	object=o2;
	ballIsInvolved = true;
  }else if(dGeomGetClass(o2)==dSphereClass){
	object=o1;
	ballIsInvolved = true;
  }
  if(object!=0){
	m_sound->playCollisionSound(dGeomGetClass(object));
  }else{
	  object = o1;
  }


  dContact contact[1];   // up to MAX_CONTACTS contacts per box-box
  for (i=0; i<1; i++) {
   
	contact[i].surface.mode = dContactBounce;
	contact[i].surface.mu = 0.10;
    contact[i].surface.mu2 = 0;
	//if(dGeomGetClass(object)==dFirstUserClass){		// HERE THE BUMPER IS HIT!!!!!!!!
	if((dGeomGetClass(object)==dCCylinderClass)&&ballIsInvolved){		// HERE THE BUMPER IS HIT!!!!!!!!
		increaseBumperHits();
		contact[i].surface.bounce = 1;
		contact[i].surface.bounce_vel = 0.01;
	}else if(dGeomGetClass(object)==dPlaneClass){	// here a plane is hit
		contact[i].surface.mode = dContactApprox1;
		contact[i].surface.bounce = 0.2;
		contact[i].surface.bounce_vel = 0.2;
	}else{	// here something other is hit
		contact[i].surface.mode = dContactBounce;
		contact[i].surface.bounce = 0.1;
		contact[i].surface.bounce_vel = 0.2;
	}
    
  }
	if (int numc = dCollide (o1,o2,MAX_CONTACTS,&contact[0].geom, sizeof(dContact))) 
	{   
	//	if(Game::Instance().onCollision(b1, b2)==true)
	//	{
			dMatrix3 RI;
			dRSetIdentity (RI);
			const dReal ss[3] = {0.02,0.02,0.02};
			for (i=0; i<numc; i++) 
			{
				dJointID c = dJointCreateContact (m_world,m_contactgroup,contact+i);
				dJointAttach (c,b1,b2);			
			}
	//	}
		
	}
  }
	
}