// AR-Praktikum ss2005 - Basti Grembowietz
// Tracker-Implementation (currently tracking 4 different markers)

#include "MarkerTracker.h"

#include <fstream> // for reading from files...
#include <sstream> // ... and parsing the input
#include <exception>

#include "PoseEstimation.h"



/// @note inherited these from PoseEstimation.h
	//! @brief Use only this constant for accessing a quaternion's X component
	static const unsigned char QX = 0;
	//! @brief Use only this constant for accessing a quaternion's Y component
	static const unsigned char QY = 1;
	//! @brief Use only this constant for accessing a quaternion's Z component
	static const unsigned char QZ = 2;
	//! @brief Use only this constant for accessing a quaternion's scalar component
	static const unsigned char QW = 3;


float MarkerTracker::getSubPixXinStripe(int ym1, int y0, int y1, int currentPositionInStripe)
{
	float a = static_cast<float>((y1 + ym1 -2*y0))/2.0f;
	float b = static_cast<float>((y1 - ym1))/2.0f;
	float c = static_cast<float>(y0);

	float coordinate = 0;

	if (a == 0.0f) // no epsilon test yet
		throw std::exception("Could not find maximum in corner-stripe");
	else
	{
		coordinate = -b / (2 * a);
		coordinate += currentPositionInStripe;
	}

	return coordinate;
}


// sobelize one stripe
void MarkerTracker::sobelizeIt(int * sobels, int stripeLength, uchar* edgeColors)
{
	// -1 0 +1
	// -2 0 +2  is the sobel pattern applied on all values.
	// -1 0 +1
	for (int i = 1; i< (stripeLength-1); i++)
	{
		sobels[i-1] = edgeColors[(i+1)] + 2* edgeColors[(i+1) + stripeLength] + edgeColors[(i+1) + 2*stripeLength];
		sobels[i-1] -= edgeColors[(i-1)] + 2* edgeColors[(i-1) + stripeLength] + edgeColors[(i-1) + 2*stripeLength];
	}

}

int MarkerTracker::minStripeIndex(int * sobels, int sobelLength)
{
	int minFound = 1;
	for (int i =0 ; i< sobelLength; i++)
		if (sobels[i]<sobels[minFound])
            minFound = i;
	return minFound;
}

CvPoint3D32f MarkerTracker::homoLineFromPointDirection(CvPoint2D32f direction, CvPoint2D32f point)
{
	CvPoint3D32f result;

	result.x = direction.y;
	result.y = -direction.x;
	result.z = point.y * direction.x - point.x * direction.y;

	return result;
}

// lines are given by normalized direction vx + one point px on it.
bool MarkerTracker::calcIntersect2D(CvPoint2D32f p0, CvPoint2D32f v0, CvPoint2D32f p1, CvPoint2D32f v1, CvPoint2D32f & intersection)
{
	bool linesIntersect = false;

	// plan: 1) homogenize it
	//       2) build cross product
	//       3) if z == 0 there is no intersection, otherwise scale it and return.

	// ok, let's homogenize:
	CvPoint3D32f line0 = homoLineFromPointDirection(v0, p0);
	CvPoint3D32f line1 = homoLineFromPointDirection(v1, p1);

	// cross product:
	CvPoint3D32f intersectionPoint;
	intersectionPoint.x = line0.y*line1.z - line0.z*line1.y;
	intersectionPoint.y = line0.z*line1.x - line0.x*line1.z;
	intersectionPoint.z = line0.x*line1.y - line0.y*line1.x;

	if (intersectionPoint.z != 0.0f)
	{
		// lines do intersect
		linesIntersect = true;

		// now rescale and write (assuming intersection != NULL)
		intersection.x = intersectionPoint.x / intersectionPoint.z;
		intersection.y = intersectionPoint.y / intersectionPoint.z;
		
	}
	return linesIntersect;
}


bool MarkerTracker::getMarkerEdges(CvSeq * quadrangleSeq, CvPoint2D32f * marker)
{
	bool success = true;

	// plan: 1. get points
	//       2. determine 8 points per line
	//       3. get 3xX pixels of these points
	//       4. apply sobel
	//       5. be happy

	// 1. extract points from sequences:
	//  a. create enough space
	CvPoint* pointArray = NULL;
	pointArray = new CvPoint [quadrangleSeq->total];
	//  b. copy points there
	cvCvtSeqToArray(quadrangleSeq, pointArray, CV_WHOLE_SEQ);

	// 1st step: determine 8 points on a line;
	//  this will first be used for calculation approximate values on the lines,
	//  after sobelization this holds the exact points on the lines.
	CvPoint2D32f pointsOnLine[8*4];
	int pointIntensity[8*4];

	// calculate directions of 4 edges
	CvPoint2D32f direction[4];

	// use size of the contour for length of the stripes
	double perimeter = cvContourPerimeter(quadrangleSeq);
	int stripeLength = static_cast<int>(perimeter / 15.0);

	// adjust stripeLength
	{
		// must not be even
		if (stripeLength %2 == 0)
			stripeLength++;
		// and at least 5
		if (stripeLength < 5)
			stripeLength = 5;
	}
	// get space for sobel: 3xstripLength is size of one stripe
	//   8 stripes, 4 edges
	unsigned char * edgeColors = NULL;
	edgeColors = new unsigned char [stripeLength * 3 * 8 * 4];

	// 2) create temporary image to extract the subpixels to		
	IplImage* edgyPic = cvCreateImage( cvSize(stripeLength,3), m_currentGreyScaleImage->depth, m_currentGreyScaleImage->nChannels );


	// get space for sobelized values
	int * sobels = NULL;
	sobels = new int [(stripeLength-2) * 8 * 4];

	try 
	{
		for (int j = 0; j < 4; j++)
		{
			// calculate direction of the edge
			direction[j].x = static_cast<float>(pointArray[(j+1)%4].x - pointArray[j%4].x);
			direction[j].y = static_cast<float>(pointArray[(j+1)%4].y - pointArray[j%4].y);

			float length = sqrtf(direction[j].x * direction[j].x + direction[j].y * direction[j].y);

			// scale factor for determining the points on the line
			float scale = length / 9.0f;

			// normalize directional vectors
			direction[j].x /= length;
			direction[j].y /= length;

			int offset = j*8*stripeLength*3; // go to beginning of current edge

			for (int i = 0; i<8; i++)
			{
				// calculate points on line
				pointsOnLine[i + j*8].x = pointArray[j%4].x + (i+1) * direction[j].x * scale;
				pointsOnLine[i + j*8].y = pointArray[j%4].y + (i+1) * direction[j].y * scale;

				// get pixel-stripes for sobel:
				float m[6];
				CvMat M = cvMat( 2, 3, CV_32F, m );
				// 1) build up transformation matrix
				m[0] = -direction[j].y; // -y
				m[1] = -direction[j].x; // -x
				m[2] = pointsOnLine[i + j*8].x;
				m[3] = direction[j].x; //x
				m[4] = -direction[j].y; //-y
				m[5] = pointsOnLine[i + j*8].y;

				// 3) extract subpixels
				cvGetQuadrangleSubPix( m_currentGreyScaleImage, edgyPic, &M, 0, cvScalarAll(0));

				// 4) copy these pixels to safety...
				//for ( int yy = 0; yy < 3; yy++)
				//	for ( int xx = 0; xx < stripeLength; xx++)
				//		edgeColors[offset + xx + yy*stripeLength] = ((uchar*)(edgyPic->imageData + edgyPic->widthStep*yy))[xx];
				for (int yy = 0; yy < 3; yy++)
					memcpy(&edgeColors[offset + yy*stripeLength], (edgyPic->imageData + edgyPic->widthStep*yy), stripeLength);

				// determine base position in sobel field
				int sobelOffset = (stripeLength-2)*i + 8*(stripeLength-2)*j;

				// determine sobel values
				sobelizeIt(&sobels[sobelOffset], stripeLength, &edgeColors[offset]);

				offset += stripeLength*3; // go to next point on the edge

				int coordMinS = minStripeIndex(&sobels[sobelOffset], stripeLength-2);

				// save intensity of the point
				pointIntensity[i + j*8] = sobels[sobelOffset + coordMinS];
				// We use min.
				float xCoord = getSubPixXinStripe(sobels[sobelOffset + coordMinS -1], 
					sobels[sobelOffset + coordMinS],
					sobels[sobelOffset + coordMinS +1],
					coordMinS);

				float x = xCoord - (stripeLength / 2); // relativ zur Mitte
				float y = 1; // relativ zur Mitte

				// now undo the rotation...
				CvPoint2D32f onLine;
				onLine.x = m[0] * x + m[1] * y + m[2];
				onLine.y = m[3] * x + m[4] * y + m[5];

				// ... then undo the translation and assign to pointsOnLine[i]
				pointsOnLine[i + j*8].x = onLine.x;
				pointsOnLine[i + j*8].y = onLine.y;

			} // all eight steps on an edge passed.
		} // all four corners & edges visited.

		// "remove" (in this case mark with red X) possible artefacts from marker-detection
		float meanIntensity = 0.0;
		for (int i = 0; i < 4*8; i++)
			meanIntensity += pointIntensity[i] / 32.0f;
		int meanInt = static_cast<int>(abs(meanIntensity));

		// we get out here via throwing an exception.
		if (meanInt < m_minMeanIntensity)
			throw std::exception("Marker sucks (intensity too low)");

		float line[4*4]; // will be (vx, vy, x0, y0) * 4

		// fit lines with the help of our points
		for (int i = 0; i < 4; i++)
		{
			CvMat pointMat = cvMat(1, 8, CV_32FC2, &pointsOnLine[i*8]);
			cvFitLine(&pointMat, CV_DIST_L2, 0, 0.01, 0.01, &line[i*4]); 
		}
		
		// now calculate exact intersection of lines and draw these points
		for (int i = 0; i < 4; i++)
		{
			CvPoint2D32f cornerF;
			if (calcIntersect2D(cvPoint2D32f(line[i*4 +2], line[i*4 +3]),
				cvPoint2D32f(line[i*4], line[i*4 +1]),
				cvPoint2D32f(line[((i+1)*4 +2)%16], line[((i+1)*4 +3)%16]),
				cvPoint2D32f(line[((i+1)*4)%16], line[((i+1)*4 + 1)%16]),
				cornerF))
			{
				marker[i] = cornerF;
			}
		}

	}
	catch (std::exception e)
	{
		// well, do nothing right now...
		// the marker is not added, that is the most important thing.
		success = false;
		// maybe a debug-message would be nice, but not nescessary.
	}

	// clean up
	delete [] edgeColors;
	delete [] sobels;
	delete [] pointArray;

	cvReleaseImage(&edgyPic);


	return success;
}

int MarkerTracker::getMarkerCodeFromArray(char * marker)
{
 	int temp = 0;
	for (int j = 1; j<5; j++)
	{
		for (int i = 0; i<4; i++)
		{
			temp = temp << 1;
			if (marker[j*6 + i + 1] > m_blackOnMarkerThreshold)
			{
				temp += 1;
			}
		}
	}

//	std::cout << "0x" << std::hex << temp << " ";
	return temp;
}

bool MarkerTracker::calcIntersect2Dfrom4Points(CvPoint2D32f * fourPoints, CvPoint2D32f & result)
{
	CvPoint2D32f dir1, dir2;
	dir1.x = fourPoints[0].x - fourPoints[2].x;
	dir1.y = fourPoints[0].y - fourPoints[2].y;

	dir2.x = fourPoints[1].x - fourPoints[3].x;
	dir2.y = fourPoints[1].y - fourPoints[3].y;

	return calcIntersect2D(fourPoints[0], dir1, fourPoints[1], dir2, result);
}

void MarkerTracker::getPosesFromContours(CvSeq * contour, CvMemStorage * storage) // , IplImage * imageToDrawOn, IplImage * greyScaleImage)
{
	CvSeq * appC = NULL;

	CvPoint2D32f marker[4];
	CvPoint2D32f inMarkerPic[4];
	inMarkerPic[0].x = inMarkerPic[0].y = inMarkerPic[1].x = inMarkerPic[3].y = -0.5;
	inMarkerPic[2].x = inMarkerPic[2].y = inMarkerPic[1].y = inMarkerPic[3].x = 5.5;

	float transformMatrix[9];
	CvMat transMat = cvMat(3, 3, CV_32FC1, &transformMatrix[0]);
	// field for the marker-data
	char markerPoints[36];

	for (; contour != NULL; contour= contour->h_next)
	{
		// check size of the contour
		CvRect boundingBox = cvBoundingRect(contour, 0);
		if ((boundingBox.height * boundingBox.width) >= m_minSquare)
		{
			// ok, big enough
			// approximate polygonal curve
			appC = cvApproxPoly(contour, sizeof(CvContour), storage, CV_POLY_APPROX_DP, cvContourPerimeter(contour)*0.02, 0);
			// we have a quad
			if (((appC->flags & (/*CV_SEQ_FLAG_CLOSED|*/CV_SEQ_FLAG_HOLE)) != 0) && (appC->total == 4))
			{

				if (getMarkerEdges(appC, &marker[0]))
				{
					cvWarpPerspectiveQMatrix(&inMarkerPic[0], &marker[0], &transMat);  
					// draw first marker of the picture
					for (int y = 0; y<6; y++)
					{
						for (int x = 0; x<6; x++)
						{
							float currentX = (x)* transformMatrix[0] + (y)* transformMatrix[1] + (1) * transformMatrix[2];
							float currentY = (x)* transformMatrix[3] + (y)* transformMatrix[4] + (1) * transformMatrix[5];
							float upToScale = (x)* transformMatrix[6] + (y)* transformMatrix[7] + (1) * transformMatrix[8];

							// un-homogenize
							if (upToScale != 0.0f)
							{
								currentX /= upToScale;
								currentY /= upToScale;
								if ((currentX >= 0.0f) && (currentY >= 0.0f) && (currentX < m_currentGreyScaleImage->width) && (currentY < m_currentGreyScaleImage->height))
								{
									markerPoints[y*6 + x] = (m_currentGreyScaleImage->imageData + m_currentGreyScaleImage->widthStep * round(currentY))[round(currentX)];
								}
							}
						}
					}
					// see if marker has black border:
 					bool markerOkay = true;
					// check upper and lower border
					for (int i = 0; i<6; i++)
						markerOkay = markerOkay && (markerPoints[i] > m_blackOnMarkerThreshold) && (markerPoints[i + 5*6] > m_blackOnMarkerThreshold);
					if (markerOkay)
					{
						// check left and right column
						for (int i = 1; i < 5; i++)
							markerOkay = markerOkay && (markerPoints[6*i] > m_blackOnMarkerThreshold) && (markerPoints[6*i + 5] > m_blackOnMarkerThreshold);
					}
					if (markerOkay) // i.e. has a black border
					{
						// marker is alright
						int markerCode = getMarkerCodeFromArray(markerPoints);
						
						// if the marker code returned is 0, the marker is not recognized, so ignore
						if (m_markers.find(markerCode) != m_markers.end())
						{
							float rotationQuaternion[4]; float translation[3];
							
							float shiftY = m_currentImage->height / 2.0f ;
							float shiftX = m_currentImage->width / 2.0f ;
							// now shift and rotate marker coordinates to center of projection
							CvPoint2D32f currentMarker[4];

							for (int i = 0; i < 4; i++)
							{
								currentMarker[i].x = marker[(i + m_markers[markerCode].shiftIndex)%4].x - shiftX;
								currentMarker[i].y = marker[(i + m_markers[markerCode].shiftIndex)%4].y - shiftY;
							}

							// estimate pose of the marker according to marker data
							estimateSquarePose(&rotationQuaternion[0], &translation[0], &currentMarker[0], m_markers[markerCode].size);

							// construct quat-trans for marker-code:
							TransQuat transform;
							// copy quaternion
							transform.q.w = rotationQuaternion[QW];
							transform.q.x = rotationQuaternion[QX];
							transform.q.y = rotationQuaternion[QY];
							transform.q.z = rotationQuaternion[QZ];
							// copy translation
							transform.t.x = translation[0];
							transform.t.y = translation[1];
							transform.t.z = translation[2];

//							m_currentTransforms[markerCode] = transform;
							m_currentTransforms[m_markers[markerCode].uniqueID] = transform;
						} // if (markerCode)
					} // if (getMarkerEdges)
				} // if 4 edges
			} 
		}
	}
}



MarkerTracker::MarkerTracker()
{
	// initialize images
	m_currentImage = NULL;
	m_currentGreyScaleImage = NULL;
	m_currentThresholdImage = NULL;

	// initialize image source
	m_source = NULL;

	// set threshold for cvThreshold
	//m_thresholdBlackWhiteT = 85.0f;
	m_neighbourhood= 20; 
	m_miniVariance= 255; 
	m_minSquare = 500; //1000
	m_minSobel = 200;
	m_minMeanIntensity = 200;
	m_simpleThreshold = 120; // 85;
	m_blackOnMarkerThreshold = -15;

}

MarkerTracker::~MarkerTracker()
{
	releaseImages();
	m_currentTransforms.clear();
}

void MarkerTracker::releaseImages()
{
	// release images
	if (m_currentGreyScaleImage)
		cvReleaseImage(&m_currentGreyScaleImage);
	
	if (m_currentThresholdImage)
		cvReleaseImage(&m_currentThresholdImage);	
}

bool MarkerTracker::getMarkerPose(int markerCode, TransQuat &t)
{
	bool markerIsTracked = false;
	// just return the pose of a marker identified by its markercode.
	if (m_currentTransforms.find(markerCode) != m_currentTransforms.end())
	{
		markerIsTracked = true;
		t = m_currentTransforms[markerCode];
	}
	return markerIsTracked;
}

void MarkerTracker::getMarkerPoseNumber(int markerNumber, TransQuat &t, float &halfMarkerSize)
{
	stdext::hash_map<int, TransQuat>::iterator iter;
	std::pair<int, TransQuat> temp;
	iter = m_currentTransforms.begin();
	for (int i = 0; i < markerNumber; i++) iter ++; // does not want it directly 8(
	//	iter = iter + markerNumber;
	temp = *iter;
	t = temp.second;
	halfMarkerSize = m_markers[temp.first].size / 2.0f;
	
	//std::hash_map<int,float*>::iterator iter;
	//std::pair<int, float*> temp;
	//	temp = *iter;
	//	delete temp.second;
}


void MarkerTracker::setSource(ImageSource * image)
{
	// if they exist, release them.
	releaseImages();

	// save image source
	m_source = image;

	// now initialize all pictures used:
	// get first frame for initializing greyscale image
	m_currentImage = m_source->getImage();		

	// initialize greyscale image
	m_currentGreyScaleImage = cvCreateImage( cvSize(m_currentImage->width, m_currentImage->height), IPL_DEPTH_8U, 1 );
	m_currentGreyScaleImage->origin = m_currentImage->origin;
	
	// initialize fixed threshold image
	m_currentThresholdImage = cvCreateImage( cvSize(m_currentImage->width, m_currentImage->height), IPL_DEPTH_8U, 1 );
	m_currentThresholdImage->origin = m_currentImage->origin;
	
}

void MarkerTracker::update()
{
	// calculate new poses from new image source.

	// get image from source.

	m_currentImage = m_source->getImage();		

	// clear all poses found in last iteration
	m_currentTransforms.clear();

	// convert image to greyscale
	cvConvertImage( m_currentImage, m_currentGreyScaleImage, 0); // do not flip.
	
	// apply thresholds
	cvThreshold(m_currentGreyScaleImage, m_currentThresholdImage, m_simpleThreshold, 255.0f, CV_THRESH_BINARY); 

	//double meanFactor = static_cast<double>(miniVariance);
	//int blockSize = 3 + neighbourhood*2;
	//double meanFactorCorrected = meanFactor / 128.0f;
	//cvAdaptiveThreshold(m_currentGreyScaleImage, m_currentThresholdImage, 255.0f, CV_ADAPTIVE_THRESH_GAUSSIAN_C, CV_THRESH_BINARY, blockSize, meanFactorCorrected); 

	// get storage for contours.
	CvMemStorage* storage = cvCreateMemStorage(0);
	
	// initialize pointers for contours.
	CvSeq* contours = NULL;

    // find contours in the images
	int CN = cvFindContours( m_currentThresholdImage, storage, &contours, sizeof(CvContour), CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint(0,0) );

//	int adaptiveCN = cvFindContours( adaptiveThreshImage, adaptiveStorage, &adaptiveContours, sizeof(CvContour), CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint(0,0) );

	// examine image from normal thresholded image
	getPosesFromContours(contours, storage);

	// release storage
	cvReleaseMemStorage(&storage);

}

bool MarkerTracker::calcIntraMarkerTransform(int markerCodeFROM, int markerCodeTO, TransQuat &transform)
{
	// check whether both markers are acquired
	TransQuat t1, t2;
	bool foundBoth = false;
	if (getMarkerPose(markerCodeFROM, t1) && getMarkerPose(markerCodeTO, t2))
	{
		foundBoth = true;
		// plan: invert t1 and go to t2

		TransQuat t1I = math3d::inverse(t1);

		// now multiply t1I and t2 and return this...
		transform = math3d::multiplyTransforms(t2, t1I);
	}
	return foundBoth;
}

int MarkerTracker::incThreshold()
{
	m_simpleThreshold += 10; 
	m_simpleThreshold %= 256; 
	return m_simpleThreshold;
}

int MarkerTracker::decThreshold()
{
	m_simpleThreshold -= 10;
	if (m_simpleThreshold < 0) 
		m_simpleThreshold = 0;
	m_simpleThreshold %= 256;
	return m_simpleThreshold;
}

int MarkerTracker::markersTracked()
{ 
	return static_cast<int>(m_currentTransforms.size());
}

bool MarkerTracker::markerIsTracked(int markerCode)
{
	return (m_currentTransforms.find(markerCode) != m_currentTransforms.end());	
}

int MarkerTracker::rotateMarkerCode(int markerCode, int numQuadrants)
{
	// more than 3 rotations are nonsense:
	numQuadrants %= 4;
	int rotatedMarkerCode = markerCode;
	for (int i = 0; i< numQuadrants; i++)
	{
		int nibble = 0;
		int tmpMarkerCode = 0;

		// generate rotated markerCode
		for (int j = 0; j < 4; j++)
		{
			// initialize bitmaps
			int mask1 = 0x0001 << j;
			int mask2 = 0x0010 << j;
			int mask3 = 0x0100 << j;
			int mask4 = 0x1000 << j;

			// generate numbers
			nibble = ((rotatedMarkerCode & mask1) << (3 - j)) | ((rotatedMarkerCode & mask2) >> (2 + j)) | ((rotatedMarkerCode & mask3) >> (7 + j)) | ((rotatedMarkerCode & mask4) >> (12 + j));
			tmpMarkerCode |= (nibble << (j * 4));
		}

		rotatedMarkerCode = tmpMarkerCode;
	}
	return rotatedMarkerCode;
}

void MarkerTracker::readMarkerConfig(std::string markerConfig)
{
	std::ifstream file(markerConfig.c_str());
	std::string line;
	std::stringstream sstr;
	if (file.is_open())
	{
		while(getline(file, line))
		{
			// format is "0x2424#0.0354234;" for each marker
			size_t start = 0;
			size_t end = 0;
			end = line.find("#", start);

			// extract marker code...
			int markerCode = 0;
			sstr << std::hex << line.substr(start, (end - start));
			sstr >> markerCode;
			sstr.clear();
			if (!markerCode)
				throw std::exception("readMarkerConfig: Could not read marker code!");

			// extract marker size
			float markerSize = 0.0f;
			start = end + 1;
			end = line.find(";", start);
			sstr << std::scientific << line.substr(start, (end - start));
			sstr >> markerSize;
			sstr.clear();
			if (markerSize == 0.0f)
				throw std::exception("readMarkerConfig: Could not read marker size!");

			// set defaults for all 4 different markers
			MarkerData md;
			md.size = markerSize;
			md.uniqueID = markerCode;
			md.shiftIndex = 0;
			// ... and put everything into the hash map. very nice.
			for (int i = 0; i < 4; i++)
			{
				int currentMarkerCode = rotateMarkerCode(markerCode, i);
				md.shiftIndex = (4 - i) %4;
				m_markers[currentMarkerCode] = md;
			}

		}
	}
	else
		throw std::exception("readMarkerConfig: Could not open file!");

}