// AR-Praktikum ss2005 - Basti Grembowietz
// this class is used to separate tracking from visualization
#pragma once

#include <hash_map>
#include <string>

#include "ImageSource.h"
#include "math3d.h"
/** @brief All tracking dependent parts go in here.
 * \ingroup AR Praktikum
 *
 *
 * \par requirements
 * win98 or later\n
 * win2k or later\n
 * OpenCV beta 4 or later\n
 * MFC\n
 *
 * \version 1.0
 * first version
 *
 * \date 05-29-2005
 *
 * \author Basti Grembowietz
 *
 * \par license
 * This code is mine. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * @todo create structs for quaternions and vectors and both. Then use this struct for hash-map and that's it.
 */
class MarkerTracker
{
private:
	IplImage * m_currentImage;
	IplImage * m_currentGreyScaleImage;
	IplImage * m_currentThresholdImage;

	ImageSource * m_source;
	// various parameters for thresholding
	int m_neighbourhood;
	int m_miniVariance;
	int m_minSquare;
	int m_minSobel;
	int m_minMeanIntensity;
	int m_simpleThreshold;
	char m_blackOnMarkerThreshold;

	/** @brief The transforms are saved with according marker code*/
	stdext::hash_map <int, TransQuat> m_currentTransforms;
	/** @brief Map (rotated markerCode)->uniqueMarkerCode, size and shiftIndex. */
	stdext::hash_map <int, MarkerData> m_markers;

	/** @brief Release acquired images
	*/
	void releaseImages();

	/** @brief Detect edge in sobelStripe with sub-pixel accuracy
	*
	* This detects the extremum that is located between three given y-values of the
	* positions x= -1, x=0 and x=1. It returns the x-coordinate of the extremum.
	* @param ym1, y0, y1 Y-Value for of x = -1, x = 0 and x = 1 respectively
	* @param currentPositionInStripe The offset of the x=0 coordnate in the stripe
	* @return x coordinate (in stripe-cos) of the extremum 
	*/
	float getSubPixXinStripe(int ym1, int y0, int y1, int currentPositionInStripe);

	/** @brief Apply the sobel operator
	* @param[out] sobels Resulting sobel values will be written here
	* @param stripeLength Length of the input stripe
	* @param edgeColors The [3 x stripelength] array of greyscale values
	*/
	void sobelizeIt(int * sobels, int stripeLength, uchar* edgeColors);


	/** @brief A standard round function
	* @param x float value to round and convert to an int
	* @return The rounded value
	*/
	int round (float x) {return static_cast<int>(x + 0.5); }

	/** @brief Acquire poses of markers from contours 
	* @param contour Contours found
	* @param storage Storage to use for sobel
	*/
	void getPosesFromContours(CvSeq * contour, CvMemStorage * storage);

	/** @brief Get subpixel accurate edges of quadrangle
	* @param quadrangleSeq Sequence of points of the quadrangle
	* @param[out] marker Subpixel accurate edges of the marker
	* @return True if successfull, false otherwise
	*/
	bool getMarkerEdges(CvSeq * quadrangleSeq, CvPoint2D32f * marker);

	/** @brief Extract 4Hex-Marker-Code of an array of greyscale values
	* @param marker The greyscale values (6x6) of the marker
	* @return HexCode representing the marker
	*/
	int getMarkerCodeFromArray(char * marker);

	/** @brief Calculate if two lines in 2D intersect and if yes, where
	* @param p0,v0 point and normalized direction vector of the first line
	* @param p1,v1 point and normalized direction vector of the second line
	* @param[out] intersection Point where the two lines meet (if they do)
	* @return True if lines intersect, false otherwise.
	*/
	bool calcIntersect2D(CvPoint2D32f p0, CvPoint2D32f v0, CvPoint2D32f p1, CvPoint2D32f v1, CvPoint2D32f & intersection);

	/** @brief Calculate the center of a quadrangle 
	*
	* Use this to calculate the center point of a quadrangle. You have to specify the four corners 
	* in order, the 1st and 3rd point will form a line as will the 2nd and 4th. Their intersection
	* point is the center of the quadrangle.
	* @param fourPoints 4 Points, the corners of the quadrangle
	* @param[out] result Center of the quadrangle (if exists)
	* @return True if center exists, false otherwise.
	*/
	bool calcIntersect2Dfrom4Points(CvPoint2D32f * fourPoints, CvPoint2D32f & result);
	
	/** @brief Represent line given by point and direction as homogeneous vector.
	* @param direction Direction of the line (needs not to be normalized)
	* @param point Point on the line
	* @return Homogeneous representation of the line
	*/
	CvPoint3D32f homoLineFromPointDirection(CvPoint2D32f direction, CvPoint2D32f point);

	/** @brief Get index of minimal element of the sobel values
	* @param sobels Array with sobel values
	* @param sobelLength Length of the array
	* @return Index of the minimal element
	*/
	int minStripeIndex(int * sobels, int sobelLength);

	/** @brief Rotate marker-code
	* @param markerCode Code of the marker
	* @param numQuadrants Number of Quadrants the code should be rotated
	* @return New marker code (what the marker would be named if rotated 90 * numQuadrants
	*/
	int rotateMarkerCode(int markerCode, int numQuadrants);

public:

	/** @brief Read marker configuration from file
	* @param markerConfig Full path to configuration file
	* @note If errors occur, exceptions are thrown.
	*/
	void readMarkerConfig(std::string markerConfig);

	MarkerTracker();
	~MarkerTracker();

	/** @brief Increase current threshold (i.e. black is darker)
	*
	* To get results more indepentendly from light-sources...
	* @return Threshold value used.
	*/
	int incThreshold();

	/** @brief Decrease current threshold (i.e. black is brighter)
	*
	* To get results more indepentendly from light-sources...
	* @return Threshold value used.
	*/
	int decThreshold();

	/** @brief Get number of markers / COS tracked
	* @return Number of markers tracked.
	*/
	int markersTracked();

	/** @brief Get world-to-marker transform (iterator-like)
	* @param markerNumber Number of the marker (no order implied)
	* @param[out] t Transform (Quaternion and Translation).
	* @param[out] halfMarkerSize Half (!) size of the marker [m]
	* @sa markersTracked() for number of markers
	*/
	void getMarkerPoseNumber(int markerNumber, TransQuat &t, float &halfMarkerSize);

	/** @brief Check if a certain marker is tracked
	* @param markerCode biggest possible code of the marker
	* @return True if marker is tracked, false otherwise.
	*/
	bool markerIsTracked(int markerCode);

	/** @brief Set the source of the image for tracker
	*
	* This is needed only once to initialize the tracker.
	* @param image image source used to acquire poses
	*/
	void setSource(ImageSource * image);

	/** @brief Calculate new poses of the markers */
	void update();

	/** @brief Get world-to-marker transform
	* @param markerCode Code of the marker (hex-code representing the layout)
	* @param[out] t Transform (Quaternion and Translation).
	* @return true if pose of the marker is acquired, false otherwise.
	*/
	bool getMarkerPose(int markerCode, TransQuat &t);

	/** @brief Calculate inter-marker transforms
	* @param markerCodeFROM Source of the transform
	* @param markerCodeTO Destination of the transform
	* @param[out] transform Resulting inter-marker transform (quaternion + vector)
	* @return true if both markers are aquired and a transform is returned.
	*/
	bool calcIntraMarkerTransform(int markerCodeFROM, int markerCodeTO, TransQuat &transform);

};