#define _USE_MATH_DEFINES 
#include <math.h>
#include <cv.h>
#include <cxcore.h>
#include <iostream>
#include <vector>


namespace 
	{
	//! @brief Use only this constant for accessing a quaternion's X component
	const unsigned char QX = 0;
	//! @brief Use only this constant for accessing a quaternion's Y component
	const unsigned char QY = 1;
	//! @brief Use only this constant for accessing a quaternion's Z component
	const unsigned char QZ = 2;
	//! @brief Use only this constant for accessing a quaternion's scalar component
	const unsigned char QW = 3;
	}


/** 
 * normalizes a quaternion (makes it a unit quaternion)
 */
float* normalizeQuaternion( float *q )
	{
	float norm = 0.0f;
	for ( int i = 0; i < 4; i++ ) 
		norm += q[i] * q[i];
	norm = sqrtf( 1.0f/ norm );
	for ( int i = 0; i < 4; i++ ) 
		q[i] *= norm;

	return q;
	}


/**
 * implementation based on proposal by Horn, idea:
 * first determine the largest entry of unit quaternion,
 * then use this to get other entries
 * adapted from dwarfutil.cpp
 */
float* matrixToQuaternion( const CvMat *pMat, float *q )
	{
	// shortcuts to the matrix data
	const float* m0 = (float*)( pMat->data.ptr );
	const float* m1 = (float*)( pMat->data.ptr + pMat->step );
	const float* m2 = (float*)( pMat->data.ptr + 2 * pMat->step );

	// get entry of q with largest absolute value
	// note: we compute here 4 * q[..]^2 - 1
	float tmp[4];
	tmp[QW] = m0[0] + m1[1] + m2[2];
	tmp[QX] = m0[0] - m1[1] - m2[2];
	tmp[QY] = -m0[0] + m1[1] - m2[2];
	tmp[QZ] = -m0[0] - m1[1] + m2[2];
	int max = QW;
	if( tmp[QX] > tmp[max] ) max = QX;
	if( tmp[QY] > tmp[max] ) max = QY;
	if( tmp[QZ] > tmp[max] ) max = QZ;

	// depending on largest entry compute the other values
	// note: these formulae can be derived very simply from the
	//       matrix representation computed in quaternionToMatrix
	switch(max) {
	case QW:
		q[QW] = sqrtf(tmp[QW]+1) * 0.5f;
		q[QX] = (m2[1] - m1[2]) / ( 4 * q[QW] );
		q[QY] = (m0[2] - m2[0]) / ( 4 * q[QW] );
		q[QZ] = (m1[0] - m0[1]) / ( 4 * q[QW] );
		break;

	case QX:
		q[QX] = sqrtf(tmp[QX]+1) * 0.5f;
		q[QW] = (m2[1] - m1[2]) / ( 4 * q[QX] );
		q[QY] = (m1[0] + m0[1]) / ( 4 * q[QX] );
		q[QZ] = (m0[2] + m2[0]) / ( 4 * q[QX] );
		break;

	case QY:
		q[QY] = sqrtf(tmp[QY]+1) * 0.5f;
		q[QW] = (m0[2] - m2[0]) / ( 4 * q[QY] );
		q[QX] = (m1[0] + m0[1]) / ( 4 * q[QY] );
		q[QZ] = (m2[1] + m1[2]) / ( 4 * q[QY] );
		break;

	case QZ:
		q[QZ] = sqrtf(tmp[QZ]+1) * 0.5f;
		q[QW] = (m1[0] - m0[1]) / ( 4 * q[QZ] );
		q[QX] = (m0[2] + m2[0]) / ( 4 * q[QZ] );
		q[QY] = (m2[1] + m1[2]) / ( 4 * q[QZ] );
		break;
	}

	normalizeQuaternion(q);

	return q;
	}


/**
 * rotate a vector around a quaternion
 * implementation based on precomputation, correct mult
 * matrix can be found in Horn, Closed-form solution of
 * absolute orientation using unit quaternions. (1987)
 * adapted from dwarfutil.cpp
 */
float* rotateQuaternion( float *r, const float *q, const float *p )
	{
	// precomputation of some values
	float xy = q[QX]*q[QY];
	float xz = q[QX]*q[QZ];
	float yz = q[QY]*q[QZ];
	float ww = q[QW]*q[QW];
	float wx = q[QW]*q[QX];
	float wy = q[QW]*q[QY];
	float wz = q[QW]*q[QZ];

	r[0] = p[0] * ( 2*(q[QX]*q[QX] + ww) - 1 ) + p[1] * 2 * (xy - wz) + p[2] * 2 * (wy + xz);
	r[1] = p[0] * 2 * (xy + wz) + p[1] * ( 2*(q[QY]*q[QY] + ww) - 1 ) + p[2] * 2 * (yz - wx);
	r[2] = p[0] * 2 * (xz - wy) + p[1] * 2 * (wx + yz) + p[2] * ( 2*(q[QZ]*q[QZ] + ww) - 1 );

	return r;
	}


/** 
 * computes the orientation and translation of a square using homography
 * @param pRot result as quaternion
 * @param pTrans result position
 * @param p2D four input coordinates. the origin is assumed to be at the camera's center of projection
 * @param fMarkerSize side-length of marker. Origin is at marker center.
 * @param f focal length
 */
void poseFromHomography( float* pRot, float *pTrans, const CvPoint2D32f* p2D, float fMarkerSize, float f )
	{
	// compute homography
	float fCp = fMarkerSize / 2.0f;
	CvPoint2D32f SrcPoints[ 4 ] = 
		{ { -fCp, fCp }, { -fCp, -fCp }, { fCp, -fCp }, { fCp, fCp } };
	CvMat* pHomography = cvCreateMat( 3, 3, CV_32F );
	cvWarpPerspectiveQMatrix( SrcPoints, p2D, pHomography );
	
	// compute inverse camera matrix 
	float fTx = 0.0f;
	float fTy = 0.0f;
	float fInvY = 1.0f; // invert y axis because of image orientation?
	float fCamInv[3][3] = 
		{ { 1.0f / f, 0.0f, -fTx / f }, { 0.0f, fInvY * 1.0f / f, fInvY * -fTy / f }, { 0.0f, 0.0f, -1.0f } };
	CvMat CamInv = cvMat( 3, 3, CV_32F, fCamInv );
	
	// compute rotation matrix
	CvMat* pRotMat = cvCreateMat( 3, 3, CV_32F );
	cvMatMul( &CamInv, pHomography, pRotMat );

	// get shortcuts for columns
	CvMat ColX;
	CvMat ColY;
	CvMat ColZ;
	cvGetCol( pRotMat, &ColX, 0 ); 
	cvGetCol( pRotMat, &ColY, 1 ); 
	cvGetCol( pRotMat, &ColZ, 2 ); 

	// compute length of the first two colums
	float fXLen = static_cast<float>( cvNorm( &ColX ) );
	float fYLen = static_cast<float>( cvNorm( &ColY ) );
	
	// copy & normalize translation
	float fTransScale = 2.0f / ( fXLen + fYLen );
	for ( int i = 0; i < 3; i++ )
		pTrans[ i ] = ((float*)( pRotMat->data.ptr + pRotMat->step * i ))[ 2 ] * fTransScale;
		
	// normalize first two colums
	cvScale( &ColX, &ColX, 1.0f / fXLen );
	cvScale( &ColY, &ColY, 1.0f / fYLen );
	
	// compute error = cosine of angle between colums
	// float fDotProd = cvDotProduct( &ColX, &ColY );

	// compute third row as vector product
	cvCrossProduct( &ColX, &ColY, &ColZ );
	
	// normalize cross product	
	float fZLen = static_cast<float>( cvNorm( &ColZ ) );
	cvScale( &ColZ, &ColZ, 1.0f / fZLen );
	
	// recompute y vector from x and z
	cvCrossProduct( &ColX, &ColZ, &ColY );
	cvScale( &ColY, &ColY, -1.0 );

	// compute rotation quaternion from matrix
	matrixToQuaternion( pRotMat, pRot );
	
	// free up stuff
	cvReleaseMat( &pRotMat );
	cvReleaseMat( &pHomography );
	}


/**
 * computes where a point should be on the screen, given a pose
 * @param p2D output 2d vector
 * @param p3d input 3D vector
 * @param pRotation rotation as quaternione
 * @param pTranslation 3-element translation
 * @param f focal length
 */
void projectPoint( CvPoint2D32f& p2D, CvPoint3D32f p3D, const float* pRotation, const float* pTranslation, float f )
	{
	float point[ 3 ];
	float point3D[ 3 ];
	point3D[0] = p3D.x;
	point3D[1] = p3D.y;
	point3D[2] = p3D.z;

	// rotate
	rotateQuaternion( point, pRotation, point3D );

	// translate
	for ( int i = 0; i < 3; i++ )
		point[ i ] += pTranslation[ i ];

	// project
	// TODO: check for division by zero
	p2D.x = f * point[ 0 ] / -point[ 2 ];
	p2D.y = f * point[ 1 ] / -point[ 2 ];
	}


/**
 * factors a non-unit-length of a quaternion into the translation
 */
void normalizePose( float* pRot, float* pTrans )
	{
	// compute length of quaternion
	float fQuatLenSq = 0.0f;
	for ( int i = 0; i < 4; i++ )
		fQuatLenSq += pRot[ i ] * pRot[ i ];
	float fQuatLen = sqrtf( fQuatLenSq );

	// normalize quaternion
	for ( int i = 0; i < 4; i++ )
		pRot[ i ] /= fQuatLen;
	
	// scale translation
	for ( int i = 0; i < 3; i++ )
		pTrans[ i ] /= fQuatLenSq;
	}


/**
 * computes the Jacobian for optimizing the pose
 * @param pResult 2x7 matrix
 * @param pParam rotation parameters: 4 * quaternion rotation + 3 * translation
 * @param p3D the 3d input vector
 * @param f focal length
 */
void computeJacobian( float* pResult, float* pParam, const CvPoint3D32f& p3D, float f )
	{
	// TODO: check for division by zero
	// maple-generated code
	float t4 = pParam[0]*p3D.x+pParam[1]*p3D.y+pParam[2]*p3D.z;
	float t10 = pParam[3]*p3D.x+pParam[1]*p3D.z-pParam[2]*p3D.y;
	float t15 = pParam[3]*p3D.y-pParam[0]*p3D.z+pParam[2]*p3D.x;
	float t20 = pParam[3]*p3D.z+pParam[0]*p3D.y-pParam[1]*p3D.x;
	float t22 = -t4*pParam[2]+t10*pParam[1]-t15*pParam[0]-t20*pParam[3]-pParam[6];
	float t23 = 1.0f/t22;
	float t24 = 2.0f*f*t4*t23;
	float t30 = f*(t4*pParam[0]+t10*pParam[3]-t15*pParam[2]+t20*pParam[1]+pParam[4]);
	float t31 = t22*t22;
	float t32 = 1.0f/t31;
	float t33 = -2.0f*t32*t15;
	float t38 = 2.0f*t32*t10;
	float t43 = -2.0f*t32*t4;
	float t47 = 2.0f*f*t10*t23;
	float t48 = -2.0f*t32*t20;
	float t51 = f*t23;
	float t60 = f*(t4*pParam[1]+t10*pParam[2]+t15*pParam[3]-t20*pParam[0]+pParam[5]);
	pResult[0] = t24-t30*t33;
	pResult[1] = 2.0f*f*t20*t23-t30*t38;
	pResult[2] = -2.0f*f*t15*t23-t30*t43;
	pResult[3] = t47-t30*t48;
	pResult[4] = t51;
	pResult[5] = 0.0f;
	pResult[6] = t30*t32;
	pResult[7+0] = -2.0f*f*t20*t23-t60*t33;
	pResult[7+1] = t24-t60*t38;
	pResult[7+2] = t47-t60*t43;
	pResult[7+3] = 2.0f*f*t15*t23-t60*t48;
	pResult[7+4] = 0.0f;
	pResult[7+5] = t51;
	pResult[7+6] = t60*t32;
	}


/**
 * computes the reprojection error for each point
 * @param pError output: two entries (x,y) for each input point = (measured - reprojected)
 * @param p3D 3D coordinates of the points
 * @param p2D measured 2D coordinates
 * @param nPoints number of input points
 * @param pRot rotation
 * @param pTrans translation
 * @param f focal length
 * @returns absolute squared error
 */
float computeReprojectionError( float* pError, const CvPoint3D32f* p3D, const CvPoint2D32f* p2D, int nPoints, 
	const float* pRot, const float* pTrans, float f )
	{
	float fAbsErrSq = 0.0f;
	
	for ( int i = 0; i < nPoints; i++ )
		{
		// reproject
		CvPoint2D32f projected;
		projectPoint( projected, p3D[ i ], pRot, pTrans, f );

		// compute deviation
		pError[ 2 * i ]     = p2D[ i ].x - projected.x;
		pError[ 2 * i + 1 ] = p2D[ i ].y - projected.y;

		// update absolute error
		fAbsErrSq += pError[ 2*i ] * pError[ 2*i ] + pError[ 2*i + 1 ] * pError[ 2*i + 1 ];
		}
	
	return fAbsErrSq;
	}


/**
 * optimize a pose with levenberg-marquardt
 * @param pRotation rotation as quaternion, both used as output and initial value
 * @param pTranslation 3-element translation, both used as output and initial value
 * @param nPoints number of correspondences
 * @param p2D pointer to camera coordinates
 * @param p3D pointer to object coordinates
 * @param f focal length
 */
void optimizePose( float* pRotation, float* pTranslation, int nPoints, const CvPoint2D32f* p2D, const CvPoint3D32f* p3D, float f )
	{
	using std::vector;
	
	float params[ 7 ];
	// copy rot & trans to vector
	for ( int i = 0; i < 4; i++ )
		params[ i ] = pRotation[ i ];
	for ( int i = 0; i < 3; i++ )
		params[ i + 4 ] = pTranslation[ i ];

	float paramDiff[ 7 ];
	vector< CvPoint2D32f > estMeasurements( nPoints );
	vector< float > jacobian( nPoints * 2 * 7 );
	vector< float > measurementDiffPrev( nPoints * 2 );
	vector< float > measurementDiffNew( nPoints * 2 );
	float MDiff2[ 7 ];
	float jacobiSquare[ 7 * 7 ];
	float fLambda = 1.0f; // levenberg-marquardt-lambda
	
	CvMat matJacobian;
	cvInitMatHeader( &matJacobian, nPoints * 2, 7, CV_32F, &jacobian[ 0 ] );
	CvMat matJacobiSquare;
	cvInitMatHeader( &matJacobiSquare, 7, 7, CV_32F, &jacobiSquare[ 0 ] );
	CvMat matMeasurementDiffPrev;
	cvInitMatHeader( &matMeasurementDiffPrev, nPoints * 2, 1, CV_32F, &measurementDiffPrev[ 0 ] );
	CvMat matMeasurementDiffNew;
	cvInitMatHeader( &matMeasurementDiffNew, nPoints * 2, 1, CV_32F, &measurementDiffNew[ 0 ] );
	CvMat matParamDiff;
	cvInitMatHeader( &matParamDiff, 7, 1, CV_32F, &paramDiff[ 0 ] );
	CvMat matMDiff2;
	cvInitMatHeader( &matMDiff2, 7, 1, CV_32F, &MDiff2[ 0 ] );
	
	// compute initial error
	float fPreviousErr = computeReprojectionError( &measurementDiffPrev[ 0 ], p3D, p2D, nPoints, params, params + 4, f );
	
	// debugging
	//std::cout << "initial error: " << fPreviousErr << std::endl;
	
	// iterate (levenberg-marquardt)
	const int nMaxIterations = 3;
	for ( int iIteration = 0; iIteration < nMaxIterations; iIteration++ )
		{
		// create jacobian
		for ( int i = 0; i < nPoints; i++ )
			computeJacobian( &jacobian[ i * 2 * 7 ], params, p3D[ i ], f );

		// multiply both sides with J^T
		cvMulTransposed( &matJacobian, &matJacobiSquare, 1 );
		cvGEMM( &matJacobian, &matMeasurementDiffPrev, 1.0, NULL, 1.0, &matMDiff2, CV_GEMM_A_T );
		
		// add lambda to diagonal
		for ( int i = 0; i < 7; i++ )
			jacobiSquare[ i * 7 + i ] += fLambda;		
		
		// do least squares
		cvSolve( &matJacobiSquare, &matMDiff2, &matParamDiff, CV_SVD );

		// update parameters		
		float paramsNew[ 7 ];
		for ( int i = 0; i < 7; i++ )
			paramsNew[ i ] = params[ i ] + paramDiff[ i ];

		// factor the quaternion length into the translation
		normalizePose( paramsNew, paramsNew + 4 );

		// compute new error
		float fErr = computeReprojectionError( &measurementDiffNew[ 0 ], p3D, p2D, nPoints, paramsNew, paramsNew + 4, f );
	
		// debugging
		//std::cout << "Iteration: " << iIteration;

		if ( fErr >= fPreviousErr )
			fLambda *= 10.0f;
		else
			{
			fLambda /= 10.0f;
			
			// update parameters
			for ( int i = 0; i < 7; i++ )
				params[ i ] = paramsNew[ i ];

			// copy measurement error
			for ( int i = 0; i < nPoints * 2; i++ )
				measurementDiffPrev[ i ] = measurementDiffNew[ i ];

			fPreviousErr = fErr;
			}
			
		// more debugging
		//std::cout << " fErr=" << fErr << " lambda=" << fLambda << std::endl;
		}

	// copy back rot & trans fromvector
	for ( int i = 0; i < 4; i++ )
		pRotation[ i ] = params[ i ];
	for ( int i = 0; i < 3; i++ )
		pTranslation[ i ] = params[ i + 4 ];
	}


/** 
 * @param pRot result as quaternion
 * @param pTrans result position
 * @param p2D coordinates of the four corners in counter-clock-wise order. 
 *        the origin is assumed to be at the camera's center of projection
 * @param fMarkerSize side-length of marker. Origin is at marker center.
 */
void estimateSquarePose( float* pRot, float *pTrans, const CvPoint2D32f* p2D, float fMarkerSize )
	{
	// approximate focal length for logitech quickcam 4000 at 320*240 resolution
	static const float fFocalLength = 400;
	
	// compute initial pose
	poseFromHomography( pRot, pTrans, p2D, fMarkerSize, fFocalLength );

	// corner 3D coordinates
	float fCp = fMarkerSize / 2.0f;
	CvPoint3D32f points3D[ 4 ] =
		{ { -fCp, fCp, 0.0f }, { -fCp, -fCp, 0.0f }, { fCp, -fCp, 0.0f }, { fCp, fCp, 0.0f } }; // counter-clock-wise

	// refine pose using nonlinear optimization
	optimizePose( pRot, pTrans, 4, p2D, points3D, fFocalLength );
	}
