#include "math3d.h"
#include <exception>
TransHomo math3d::createMatrix( TransQuat t)
{
	// Note: This is taken from the matrix-quaternion-faq.

	// pre-computations of squares and other products:
	float xx, xy, xz, xw, yy, yz, yw, zz, zw;

	xx = t.q.x * t.q.x;
    xy = t.q.x * t.q.y;
    xz = t.q.x * t.q.z;
    xw = t.q.x * t.q.w;

	yy = t.q.y * t.q.y;
    yz = t.q.y * t.q.z;
    yw = t.q.y * t.q.w;

	zz = t.q.z * t.q.z;
    zw = t.q.z * t.q.w;

	// layout of the opengl matrix:
	/*
	m = 0  4  8 12
		1  5  9 13
		2  6 10 14  
		3  7 11 15

	*/
	TransHomo tResult;
	// ok, now compute the matrix
	tResult.t[0] = 1 - 2* (yy + zz);
	tResult.t[4] = 2 * (xy - zw);
	tResult.t[8] = 2 * (xz + yw);
	tResult.t[12] = t.t.x;

	tResult.t[1] = 2 * (xy + zw);
	tResult.t[5] = 1 - 2* (xx + zz);
	tResult.t[9] = 2 * (yz - xw);
	tResult.t[13] = t.t.y;

	tResult.t[2] = 2 * (xz - yw);
	tResult.t[6] = 2 * (yz + xw);
	tResult.t[10] = 1 - 2*(xx + yy);
	tResult.t[14] = t.t.z;

	tResult.t[3] = tResult.t[7] = tResult.t[11] = 0;
	tResult.t[15] = 1;
	
	return tResult;
}

void math3d::normalize(quaternion &q)
{
	float length = sqrt(q.x * q.x + q.y * q.y + q.z * q.z + q.w * q.w);
	if (length == 0.0f)
		throw std::exception("normalize: Quaternion has 0 length.");

	q.w /= length;
	q.x /= length;
	q.y /= length;
	q.z /= length;
}

void math3d::normalize(vector3f &v)
{
	float length = sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
	if (length == 0.0f)
		throw std::exception("normalize: Vector has 0 length.");

	v.x /= length;
	v.y /= length;
	v.z /= length;
}

TransQuat math3d::interpolateTransforms(TransQuat start, TransQuat end, float progress)
{
	TransQuat result;

	// first interpolate quaternion (SLERP)
	float wValue = start.q.w * end.q.w;
	float zValue = start.q.z * end.q.z;

	float cosOmega = start.q.w * end.q.w - start.q.x * end.q.x - start.q.y * end.q.y - start.q.z * end.q.z;
	// angle has to be adjusted somehow... grrr.. or not?
	float omega = acos(cosOmega);
	float sin1_hOmega = sin ( (1.0f - progress) * omega);
	float sinhOmega = sin ( (progress) * omega);
	
	result.q.w = start.q.w * sin1_hOmega + end.q.w * sinhOmega;
	result.q.x = start.q.x * sin1_hOmega + end.q.x * sinhOmega;
	result.q.y = start.q.y * sin1_hOmega + end.q.y * sinhOmega;
	result.q.z = start.q.z * sin1_hOmega + end.q.z * sinhOmega;

	normalize(result.q);

	// next interpolate the translation linear... this is a bit easier =)
	result.t.x = start.t.x * (1.0f - progress) + end.t.x  * progress ;
	result.t.y = start.t.y * (1.0f - progress) + end.t.y  * progress ;
	result.t.z = start.t.z * (1.0f - progress) + end.t.z  * progress ;

	return result;
}

float math3d::extractAngle(quaternion q)
{
	float angle = 0;
	bool adjustZ = false;
	if (q.w < 0.0f)
	{
		// reverse sign...
		angle = 2* acos(-q.w);
		// and possibly adjust
		if (q.z > 0.0f)
			angle = (3.1415f * 2.0f) - angle ;
	}
	else
	{
		// angle of quat has positive sign: no changes
		angle = 2* acos(q.w);
		// and possibly adjust
		if (q.z < 0.0f)
			angle = (3.1415f * 2.0f) - angle ;
	}
	return angle;
}

TransHomo math3d::calcTreeHanger(vector3f zAxis)
{
	TransHomo result;

	if (!isNullVector(zAxis))
		normalize(zAxis); // just to be sure 

	vector3f xAxis = { 1.0f, 0.0f, 0.0f };
//	vector3f yAxis = crossProduct(zAxis, xAxis); // maybe the other way round
	vector3f yAxis = crossProduct(xAxis, zAxis); // maybe the other way round
	normalize(yAxis);
	///*
	//m = 0  4  8 12
	//	1  5  9 13
	//	2  6 10 14  
	//	3  7 11 15
	//*/

	result.t[3] = result.t[7] = result.t[11] = result.t[12] = result.t[13] = result.t[14] = 0.0f;
	result.t[15] = 1.0f;

	// set rotation matrix, i.e. hanger matrix
	//result.t[0] = xAxis.x;
	//result.t[1] = xAxis.y;
	//result.t[2] = xAxis.z;

	//result.t[4] = yAxis.x;
	//result.t[5] = yAxis.y;
	//result.t[6] = yAxis.z;

	//result.t[8] = zAxis.x;
	//result.t[9] = zAxis.y;
	//result.t[10] = zAxis.z;

	result.t[0] = xAxis.x;
	result.t[4] = xAxis.y;
	result.t[8] = xAxis.z;

	result.t[1] = yAxis.x;
	result.t[5] = yAxis.y;
	result.t[9] = yAxis.z;

	result.t[2] = zAxis.x;
	result.t[6] = zAxis.y;
	result.t[10] = zAxis.z;

	return result;
}

TransHomo math3d::calcXHanger(vector3f xAxis)
{
	TransHomo result;

	normalize(xAxis);

	vector3f yAxis = { xAxis.y, -xAxis.x, xAxis.z };
	vector3f zAxis = crossProduct(xAxis, yAxis);
	normalize(zAxis);
	//vector3f zAxis = { xAxis.y * yAxis.z - xAxis.z * yAxis.y,
	//					xAxis.z * yAxis.x - xAxis.x * yAxis.z,
	//					xAxis.x * yAxis.y - xAxis.y * yAxis.x };

	/*
	m = 0  4  8 12
		1  5  9 13
		2  6 10 14  
		3  7 11 15
	*/

	result.t[3] = result.t[7] = result.t[11] = result.t[12] = result.t[13] = result.t[14] = 0.0f;
	result.t[15] = 1.0f;

	result.t[0] = xAxis.x;
	result.t[1] = xAxis.y;
	result.t[2] = xAxis.z;

	result.t[4] = yAxis.x;
	result.t[5] = yAxis.y;
	result.t[6] = yAxis.z;

	result.t[8] = zAxis.x;
	result.t[9] = zAxis.y;
	result.t[10] = zAxis.z;

	return result;
	
}

TransQuat math3d::inverse(TransQuat t)
{
		TransQuat tI;

		// inverse rotation:
		tI.q.w = t.q.w;
		tI.q.x = -t.q.x;
		tI.q.y = -t.q.y;
		tI.q.z = -t.q.z;

		// tI.t = - q* t_old q
		// so rotate:
		tI.t = rotate(tI.q, t.t);
		// now inverse sign
		tI.t.x = -tI.t.x;
		tI.t.y = -tI.t.y;
		tI.t.z = -tI.t.z;

		return tI;
}

TransHomo math3d::inverse(TransHomo t)
{
	/*
	m = 0  4  8 12
		1  5  9 13
		2  6 10 14  
		3  7 11 15
	*/
	// inverse rotation == transpose
	float tmp;
	tmp = t.t[4];
	t.t[4] = t.t[1];
	t.t[1] = tmp;

	tmp = t.t[8];
	t.t[8] = t.t[2];
	t.t[2] = tmp;

	tmp = t.t[9];
	t.t[9] = t.t[6];
	t.t[6] = tmp;

	// now apply this rotation to the translation vector
	vector3f oT; //old translation
	oT.x = t.t[12];
	oT.y = t.t[13];
	oT.z = t.t[14];
	
	t.t[12] = - (oT.x * t.t[0] + oT.y * t.t[4] + oT.z * t.t[8]);
	t.t[13] = - (oT.x * t.t[1] + oT.y * t.t[5] + oT.z * t.t[9]);
	t.t[14] = - (oT.x * t.t[2] + oT.y * t.t[6] + oT.z * t.t[10]);

	return t;
}

vector3f math3d::rotate(quaternion q, vector3f p)
{
	vector3f result;

	// precomputation of some values
	float xy = q.x*q.y;
	float xz = q.x*q.z;
	float yz = q.y*q.z;
	float ww = q.w*q.w;
	float wx = q.w*q.x;
	float wy = q.w*q.y;
	float wz = q.w*q.z;

	result.x = p.x * ( 2*(q.x*q.x + ww) - 1 ) + p.y * 2 * (xy - wz) + p.z * 2 * (wy + xz);
	result.y = p.x * 2 * (xy + wz) + p.y * ( 2*(q.y*q.y + ww) - 1 ) + p.z * 2 * (yz - wx);
	result.z = p.x * 2 * (xz - wy) + p.y * 2 * (wx + yz) + p.z * ( 2*(q.z*q.z + ww) - 1 );

	return result;
}

TransQuat math3d::multiplyTransforms(TransQuat t1, TransQuat t2)
{
	TransQuat result;

	// let's first care about the quaternions: t1 = x1, y1, z1, w1 ...
	result.q.w = t2.q.w * t1.q.w - t2.q.x * t1.q.x - t2.q.y * t1.q.y - t2.q.z * t1.q.z;
	result.q.x = t2.q.w * t1.q.x + t2.q.x * t1.q.w + t2.q.y * t1.q.z - t2.q.z * t1.q.y;
	result.q.y = t2.q.w * t1.q.y + t2.q.y * t1.q.w + t2.q.z * t1.q.x - t2.q.x * t1.q.z;
	result.q.z = t2.q.w * t1.q.z + t2.q.z * t1.q.w + t2.q.x * t1.q.y - t2.q.y * t1.q.x;

	// now rotate the t1-vector
	result.t = rotate(t2.q, t1.t);
	// and add t2.t
	result.t.x += t2.t.x;
	result.t.y += t2.t.y;
	result.t.z += t2.t.z;

	// ok, done
	return result;
}

vector3f math3d::applyTransform(TransQuat t, vector3f v)
{
	vector3f result;
	// first rotate
	result = rotate(t.q, v);
	// now translate
	result.x += t.t.x;
	result.y += t.t.y;
	result.z += t.t.z;

	return result;
}

vector3f math3d::applyTransform(TransHomo t, vector3f v)
{
	// caution: OpenGL-Layout!
	vector3f result;
	// so begin the matrix multiplication
	// layout of the opengl matrix:
	/*
	m = 0  4  8 12
		1  5  9 13
		2  6 10 14  
		3  7 11 15
	*/
	result.x = t.t[0] * v.x + t.t[4] * v.y + t.t[8] * v.z + t.t[12];
	result.y = t.t[1] * v.x + t.t[5] * v.y + t.t[9] * v.z + t.t[13];
	result.z = t.t[2] * v.x + t.t[6] * v.y + t.t[10] * v.z + t.t[14];

	// check if it is homogeneous w/o projection parameters:
	if ((t.t[3] != 0.0f) || (t.t[7] != 0.0f) || (t.t[11] != 0.0f) || (t.t[15] != 1.0f))
	{
		// resize vector
		float tmp = t.t[3] * v.x + t.t[7] * v.y + t.t[11] * v.z + t.t[15];
		result.x /= tmp;
		result.y /= tmp;
		result.z /= tmp;
	}

	return result;
}

vector3f math3d::vec3f(float x, float y, float z)
{
	vector3f result = {x, y, z};
	return result;
}

vector3f math3d::nullVector()
{
	return vec3f(0.0f, 0.0f, 0.0f);
}

vector3f math3d::crossProduct(vector3f a, vector3f b)
{
	vector3f result = { a.y * b.z - a.z * b.y,
						a.z * b.x - a.x * b.z,
						a.x * b.y - a.y * b.x };
	return result;
}
bool math3d::isNullVector(vector3f x)
{
	return ((x.x == 0.0f) && (x.y == 0.0f) && (x.z == 0.0f));
}

quaternion math3d::nullQuat()
{
	quaternion q = {0.0f, 0.0f, 0.0f, 1.0f};
	return q;
}