/// all types and some math go here.
#pragma once

#include <cxtypes.h>

/// @brief Struct for better access to a float quaternion.
struct quaternion
{
	float x,y,z; /// The x,y,z-component (axis) of a quaternion
	float w;  /// w-component (angle) of a quaternion
};

/// @brief Struct for better access to 3d float vectors
typedef CvPoint3D32f vector3f;

/// @brief Struct for better access to a quat-trans transform
struct TransQuat
{
	vector3f t;	/// translation of the transform
	quaternion q; /// rotation of the transform
};


/// @brief Struct for better managing of homogeneous transforms
struct TransHomo
{
	float t[16];
};

/// @brief Struct for storing marker-information
struct MarkerData
{
	int uniqueID;	// unique id of the marker
	float size;		// size of an edge of the marker
	int shiftIndex; // defining how cos is posed on the marker
};

/** @brief Holds all kinds of three dimensional useful methods, all static */
class math3d
{
public:
	/** @brief Create a homogenized matrix out of a quaternion and a translation vector
	* @param t Transformation using a quaterion and a translation vector.
	* @return 4x4 homogenzied matrix
	*/
	static TransHomo createMatrix( TransQuat t);

	/** @brief Normalize a quaternion
	* @param q Quaternion to normalize (in place)
	* @throws If length == 0;
	*/
	static void normalize(quaternion &q);

	/** @brief Normalize a vector
	* @param v Vector to normalize (in place)
	* @throws If length == 0;
	*/
	static void normalize(vector3f &v);

	/** @brief Interpolate between two transforms
	* @param start Transform to start with
	* @param end Transform to end
	* @param progress Between 0 and 1, indicates the ratio start:end
	* @return Interpolated Transform
	*/
	static TransQuat interpolateTransforms(TransQuat start, TransQuat end, float progress);

	/** @brief Extract angle from quaterinon
	* @param q Quaternion to extract angle from
	* @return Angle (might be in all 4 quadrants) [0;2 PI]
	*/
	static float extractAngle(quaternion q);

	/** @brief Calculate hanger matrix (hang x-axis to given vector)
	* @param xAxis new x-axis
	* @return hanger matrix
	*/
	static TransHomo calcXHanger(vector3f xAxis);

	/** @brief Calculate tree hanger matrix
	* @param zAxis new z-axis
	* @return hanger matrix
	*/
	static TransHomo calcTreeHanger(vector3f zAxis);

	/** @brief Inverse transformation
	* @param t Transformation to inverse
	* @return Inversed transformation
	*/
	static TransQuat inverse(TransQuat t);

	/** @brief Inverse transformation
	* @param t Transformation to inverse
	* @return Inversed transformation
	*/
	static TransHomo inverse(TransHomo t);

	/** @brief Rotate a point by a quaternion.
	* @param q Quaternion to apply
	* @param p Point to rotate
	* @return Rotated point.
	*/
	static vector3f rotate(quaternion q, vector3f p);
	
	/** @brief Apply two transformations t1 and t2 into one.
	* @param t1 Transform 1 (first applied)
	* @param t2 Transform 2 (second applied)
	* @return resulting transform t = t2 * t1
	*/
	static TransQuat multiplyTransforms(TransQuat t1, TransQuat t2);

	/** @brief Apply a tranformation to a vector
	* @param t transformation (quaternion + translation)
	* @param v vector
	* @return The transformed vector
	*/
	static vector3f applyTransform(TransQuat t, vector3f v);

	/** @brief Apply a tranformation to a vector
	* @param t transformation (homogeneous 4x4 matrix)
	* @param v vector
	* @return The transformed vector
	*/
	static vector3f applyTransform(TransHomo t, vector3f v);

	/** @brief Create vetor3f from given components
	* @param x, y, z Components of the vector
	* @return vector3f (x, y, z)
	*/
	static vector3f vec3f(float x, float y, float z);

	/** @brief Return null-vector (0, 0, 0)
	* @return null vector
	*/
	static vector3f nullVector();

	/** @brief Crossproduct
	* @param a, b vectors to take cross product from
	* @return a x b
	*/
	static vector3f crossProduct(vector3f a, vector3f b);

	/** @brief Check if given vector is null vector
	* @param x the vector to check
	* @return true if x = (0,0,0), false otherwise
	*/
	static bool isNullVector(vector3f x);

	/** @brief Null-Quaternion: No rotation
	* @return null quaternion
	*/
	static quaternion math3d::nullQuat();
	
};














